from .whitelist import Whitelist


class IPv4WL(Whitelist):
    """IPv4 address whitelist.
    This class stores its whitelisted IPv4 address in the ipv4addr table.
    """

    @staticmethod
    def getCreateTable():
        """Create the table to hold the addresses"""
        return "create table ipv4addr (addr char(15))"

    @staticmethod
    def storeDB(conn, ipv4addrs):
        """Overwrite the addresses in the table with the specified addresses (string) in the database."""
        connection = conn.cursor()
        vals = []
        for ipv4addr in ipv4addrs:
            vals.append((ipv4addr,))
        connection.execute("delete from ipv4addr")
        connection.executemany("insert into ipv4addr(addr) values (?)", iter(vals))
        connection.close()

    def isWhitelisted(self, conn, indicatorType, indicator):
        """
        Return whether or not the indicator of type indicatorType is whitelisted by this whitelist.
        This checks the indicator against the addresses in the table and requires an exact match to be considered
        whitelisted.
        """
        connection = conn.cursor()
        connection.execute("select addr from ipv4addr where addr=?", (indicator,))
        rec = connection.fetchone()
        connection.close()
        return rec is not None



import importlib
import io
import sys
import logging
import toml
import os

from lqmt.lqm.tool import ToolChain
from lqmt.whitelist.master import MasterWhitelist
from lqmt.lqm.logging import LQMLogging
from lqmt.lqm.sourcedir import DirectorySource
from lqmt.lqm.exceptions import ConfigurationError
from lqmt.lqm.systemconfig import SystemConfig

sys.path.append('tpl/toml')


class ToolInfo:
    """Holds info for future instantiation and instantiates tool instances
    """

    def __init__(self, toolClass, configClass):
        """Hold tool class object and config class object"""
        self._toolClass = toolClass
        self._configClass = configClass

    def create(self, configData, csvToolInfo, unhandledCSV):
        """Create a new config instance from the parameters passed in
        and then create a new tool instace using that configuration"""
        config = self._configClass(configData, csvToolInfo, unhandledCSV)
        tool = self._toolClass(config)
        return tool


class LQMToolConfig(object):
    """Holds the Configuration data for the LQMTool"""

    def __init__(self, configfile):
        self._logger = logging.getLogger("LQMT.Config")
        self._sources = []
        self._parsers = {}
        self._toolChains = []
        self._whitelist = None
        self._loggingCfg = None
        self._toolsList = []

        # load config files
        self._loadSystemConfig()
        self._loadUserConfig(configfile)
        # initialize logging
        self._initializeLogging()
        # process config files
        toolClasses = self._processSystemConfig()
        self._processUserConfig(toolClasses)
        if not self._sources:
            raise ConfigurationError("No sources specified")

    def _loadSystemConfig(self):
        """Load the System configuration file"""
        sysconf = SystemConfig()
        self._config = sysconf.getConfig()

    def _processSystemConfig(self):
        """
        Parses and initializes the system configuration
        :return: toolClasses, which is a dictionary holding classes for the tools assigned in the user configuration
        """
        self._initParserConfig(self._config)
        toolClasses = self._initToolConfig(self._config)
        return toolClasses

    def _initParserConfig(self, config):
        parsers = config['parsers']
        if 'additional_paths' in parsers:
            add_path = parsers['additional_paths']
            del parsers['additional_paths']
            sys.path.extend(add_path)
        path = None
        if 'path' in parsers:
            path = parsers['path']
            del parsers['path']
        # path info is loaded from the config files
        for key in list(parsers.keys()):
            parserinfo = parsers[key]
            if path is not None:
                sys.path.append(path)
            importlib.import_module(parserinfo['module'])
            mod = importlib.import_module(parserinfo['module'] + ".parser")
            parserClass = getattr(mod, parserinfo['parser_class'])
            parserConfig = None
            if "configs" in parserinfo:
                parserConfig = parserinfo['configs']
            if parserConfig is not None:
                parser = parserClass(parserConfig)
            else:
                parser = parserClass()
            self._parsers[parserinfo['format']] = parser
        self._logger.debug("Parsers loaded: %s" % ', '.join(self._parsers.keys()))

    def _initToolConfig(self, config):
        """
        Set up the system path for the configured tools and create the configuration data for those tools so they can
        be created later.

        Currently all tools are created. Should be updated to only initialize tools that are a part of the configuration
        """
        tooldefs = config['tools']
        usertooldefs = self._userConfig['Tools'].keys()
        usertooldefs = list(usertooldefs)
        usertooldefs.append('CSV')
        if 'path' in tooldefs:
            path = tooldefs['path']
            # add any paths specified to the system path
            sys.path.append(path)
            del tooldefs['path']
        toolClasses = {}
        for key in list(tooldefs.keys()):
            if key in usertooldefs:
                toolinfo = tooldefs[key]
                if 'additional_paths' in toolinfo:
                    # add any additional paths needed by the tool
                    add_path = toolinfo['additional_paths']
                    sys.path.extend(add_path)
                # import the tool & config modules
                mod = importlib.import_module("lqmt.tools." + toolinfo['module'] + ".tool")
                toolClass = getattr(mod, toolinfo['tool_class'])
                mod = importlib.import_module("lqmt.tools." + toolinfo['module'] + ".config")
                # get the class "object" for later creation
                cfgClass = getattr(mod, toolinfo['config_class'])
                toolClasses[key] = ToolInfo(toolClass, cfgClass)

        self._logger.debug("Tools loaded: %s" % ', '.join(usertooldefs))
        return toolClasses

    def _loadUserConfig(self, configFile):
        """
        Loads the user configuration file
        :param configFile: String that defines the path of the user configuration file.
        """

        if os.path.exists(configFile):
            cfg = open(configFile)
            topLevelConfig = toml.loads(cfg.read())
            cfg.close()
        else:
            configFile = io.StringIO(configFile)
            topLevelConfig = toml.load(configFile)
        self._userConfig = {}
        self._userConfig.update(topLevelConfig)

    def _initializeLogging(self):
        if 'Logging' in self._userConfig:
            self._loggingCfg = LQMLogging(self._userConfig['Logging'])

    def _processUserConfig(self, toolClasses):
        """Process the user-level configuration"""

        # Add any sources specified
        self._addSourcesConfig(self._userConfig)

        # If a whitelist was specified, create it
        if 'Whitelist' in self._userConfig:
            self._whitelist = MasterWhitelist(configData=self._userConfig['Whitelist'])
        # create any tools and tool chains in the top-level user config file
        globalTools = self._createTools(self._userConfig, toolClasses)
        self._createToolChains(self._userConfig, globalTools)

        # process any includes
        if 'includes' in self._userConfig:
            for incl in self._userConfig['includes']:
                cfg = open(incl)
                config = toml.loads(cfg.read())
                cfg.close()
                # add any sources specified, if any
                self._addSourcesConfig(config)
                # and create all tools/tool chains specified
                localTools = self._createTools(config, toolClasses)
                self._createToolChains(config, localTools, globalTools)

    def _createTools(self, config, toolClasses):
        if 'Tools' not in config:
            return
        tools = config['Tools']

        theseTools = {}
        for key in tools:
            if key in toolClasses:
                toolInfo = toolClasses[key]
                self._toolsList.append(key)
                for cfgData in tools[key]:
                    tool = toolInfo.create(cfgData, toolClasses['CSV'], self._config['UnprocessedCSV'])
                    tool.toolName = key
                    theseTools[tool.getName()] = tool
        return theseTools

    def _createToolChains(self, config, localTools, globalTools=None):
        if 'ToolChains' not in config:
            return

        chains = config['ToolChains']
        for chainCfg in chains:
            if 'active' in chainCfg and chainCfg['active'] == True:
                self._toolChains.append(self._createToolChain(chainCfg, localTools, globalTools))
            else:
                self._logger.info("Toolchain {0} is currently set as inactive in the user configuration. "
                                  "Tools in this toolchain will not run.".format(chainCfg['name']))

    def _createToolChain(self, chainCfg, localTools, globalTools):
        chain = []
        allEnabled = True
        for toolName in chainCfg['chain']:
            if toolName in localTools:
                tool = localTools[toolName]
            elif globalTools is not None and toolName in globalTools:
                tool = globalTools[toolName]
            else:
                raise ConfigurationError("Named tool not found: " + toolName)
            chain.append(tool)
            allEnabled = allEnabled and tool.isEnabled()
            if not tool.isEnabled():
                self._logger.error("Tool chain {0} is disabled due to tool {1} being disabled".format(chainCfg['name'],
                                                                                                      tool.getName()))

        chain = ToolChain(chain, chainCfg['name'], enabled=allEnabled)
        if allEnabled:
            self._logger.info("Created tool chain: {0}".format(chain.getName()))
        chain.printTools()
        return chain

    def _addSourcesConfig(self, config):
        if 'Source' not in config:
            return
        srcCfgs = config["Source"]

        for key in srcCfgs:
            for cfg in srcCfgs[key]:
                if key == "Directory":
                    self._sources.append(DirectorySource(cfg))

    def getSources(self):
        return self._sources

    def getParser(self, fmt):
        if fmt not in self._parsers:
            return None
        return self._parsers[fmt]

    def getToolChains(self):
        return self._toolChains

    def getWhitelist(self):
        return self._whitelist

    def getToolsList(self):
        """
        :return: Returns a list of tools that are currently active in the toolchain.
        """
        return self._toolsList



class SystemConfig:
    def __init__(self):
        self.system_config = {
            'tools': {
                'path': 'tools',
                'Checkpoint': {
                    'module': 'to_checkpoint', 'config_class': 'CheckpointConfig', 'tool_class': 'ToCheckpoint'
                },
                'PaloAlto': {
                    'module': 'to_paloalto', 'config_class': 'PaloAltoConfig', 'tool_class': 'ToPaloAlto',
                    'additional_paths': ['tpl']
                },
                'CSV': {
                    'module': 'to_csv', 'config_class': 'CSVConfig', 'tool_class': 'ToCSV'
                },
                'CEF': {
                    'module': 'to_cef', 'config_class': 'CEFConfig', 'tool_class': 'ToCEF'
                },
                'ArcSight': {
                    'module': 'to_arcsight', 'config_class': 'ArcSightConfig', 'tool_class': 'ToArcSight'
                },
                'SysLog': {
                    'module': 'to_syslog', 'config_class': 'SysLogConfig', 'tool_class': 'ToSysLog'
                },
                'FlexText': {
                    'module': 'to_flextext', 'config_class': 'FlexTextConfig', 'tool_class': 'ToFlexText'
                },
                'Splunk': {
                    'module': 'to_splunk', 'config_class': 'SplunkConfig', 'tool_class': 'ToSplunk'
                },
                'Bro': {
                    'module': 'to_bro', 'config_class': 'BroConfig', 'tool_class': 'ToBro'
                }
            },
            'UnprocessedCSV': {
                'name': 'for-unprocessed',
                'fields': ['dataItemID', 'fileID', 'detectedTime', 'reportedTime', 'processedTime', 'indicator',
                           'indicatorType', 'indicatorDirection', 'secondaryIndicator', 'secondaryIndicatorType',
                           'secondaryIndicatorDirection', 'directSource', 'secondarySource', 'action1', 'duration1',
                           'action2', 'duration2', 'reason1', 'reference1', 'reason2', 'reference2', 'majorTags',
                           'minorTags', 'restriction', 'sensitivity', 'reconAllowed', 'priors', 'confidence',
                           'severity', 'relevancy', 'relatedID', 'relationType', 'comment', 'fileHasMore']
            },
            'parsers': {
                'path': 'lqm',
                'additional_paths': ['ft'],
                'cfm20': {
                    'module': 'lqmt.lqm.parsers.FlexTransform',
                    'parser_class': 'FlexTransformParser',
                    'configs': {
                        'LQMTools': 'resources/sampleConfigurations/lqmtools.cfg',
                        'Cfm20Alert': 'resources/sampleConfigurations/cfm20alert.cfg'
                    },
                    'format': 'Cfm20Alert'
                },
                'cfm13': {
                    'module': 'lqmt.lqm.parsers.FlexTransform',
                    'parser_class': 'FlexTransformParser',
                    'configs': {
                        'LQMTools': 'resources/sampleConfigurations/lqmtools.cfg',
                        'Cfm13Alert': 'resources/sampleConfigurations/cfm13.cfg'
                    },
                    'format': 'Cfm13Alert'
                },
                'stixtlp': {
                    'module': 'lqmt.lqm.parsers.FlexTransform',
                    'parser_class': 'FlexTransformParser',
                    'configs': {
                        'LQMTools': 'resources/sampleConfigurations/lqmtools.cfg',
                        'stix-tlp': 'resources/sampleConfigurations/stix_tlp.cfg'

                    },
                    'format': 'stix-tlp'
                }
            }
        }

    def getConfig(self):
        return self.system_config



from lqmt.lqm.tool import ToolConfig
from lqmt.lqm.exceptions import ConfigurationError
from lqmt.lqm.data import Alert
import logging
import os
import datetime


class CSVConfig(ToolConfig):
    def __init__(self, configData, csvToolInfo, unhandledCSV):
        super().__init__(configData, csvToolInfo, unhandledCSV)
        self._logger = logging.getLogger("LQMT.CSV.{0}".format(self.getName()))
        self.fields = ""

        hasError = False
        self._file = None
        if 'file' in configData:
            file = configData['file']
            filebase, fext = os.path.splitext(file)
            tss = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
            self._file = filebase + "." + tss + fext
        else:
            hasError = True
            self._logger.error("'file' must be specified for CSV tool")

        if 'fields' in configData:
            invalid = self._initFields(configData['fields'])
            if not invalid:
                self.fields = configData['fields']
            else:
                hasError = True
                plural = ""
                if len(invalid) > 1:
                    plural = "s"
                self._logger.error("Invalid alert field{0} specified: {1}.".format(plural, ",".join(invalid)))
        else:
            self._logger.error("fields must be specified in the configuration")
            hasError = True

        if hasError:
            self.disable()
            raise ConfigurationError("Missing a required value in the user configuration for the to_csv tool")

    def getFile(self):
        return self._file

    def _initFields(self, fields):
        invalid = []
        for fld in fields:
            if not Alert.isValidField(fld):
                invalid.append(fld)

        return invalid

    def getFields(self):
        return self.fields



import os
import logging
import datetime
from lqmt.lqm.tool import ToolConfig


class FlexTextConfig(ToolConfig):
    """
    Configuration Class for FlexText
    """

    def __init__(self, config_data, csvToolInfo, unhandledCSV):
        """
        Constructor
        """
        super().__init__(config_data, csvToolInfo, unhandledCSV)

        self.logger = logging.getLogger("LQMT.FlexText.{0}".format(self.getName()))

        # FlexTransform configuration variables
        self.header_line = False
        self.increment_file = False
        self.flext_config = 'resources/sampleConfigurations/flextext.cfg'
        self.config_dict = {}
        self.config_str = ""
        self.source_configs = {
            'Cfm13Alert': 'resources/sampleConfigurations/cfm13.cfg',
            'Cfm20Alert': 'resources/sampleConfigurations/cfm20alert.cfg',
            'stix-tlp': 'resources/sampleConfigurations/stix_tlp.cfg'
        }

        # FlexText configuration variables
        self.fileParser = self.validation('fileParser', str, default="CSV")
        self.delimiter = self.validation('delimiter', str, required=True)
        self.quote_char = self.validation('quoteChar', str, required=True)
        self.escape_char = self.validation('escapeChar', str, required=True)
        self.header_line = self.validation('headerLine', bool, default=True)
        self.double_quote = self.validation('doubleQuote', bool)
        self.quote_style = self.validation('quoteStyle', str, default="none")
        self.primary_schema_config = self.validation('primarySchemaConfig', str, default="resources/schemaDefinitions"
                                                                                         "/lqmtools.json")
        self.increment_file = self.validation('incrementFile', bool)
        self.file = self.validation('fileDestination', str, required=True)

        # fields variable is configured differently for legacy purposes. The fields variable was originally a string,
        # so we are first checking to see if it's a string (legacy configs), if so we convert it to a list. Then we
        # validate it.
        if 'fields' in config_data:
            if isinstance(config_data['fields'], str):
                config_data['fields'] = config_data['fields'].split(',')

            self.fields = self.validation('fields', list, required=True)

        if self.increment_file:
            base, extension = os.path.splitext(self.file)
            file_name = "." + datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
            self.file_destination = base + file_name + extension

    def config_to_dict(self):
        """
        Takes config vars and turns them into a FlexT compatible config dict; assigns values to config_dict
        :return: returns config_dict var
        """
        self.config_dict = {
            'SYNTAX': {
                'FileParser': self.fileParser
            },
            'CSV': {
                'Fields': ','.join(self.fields),
                'Delimiter': self.delimiter,
                'QuoteChar': self.quote_char,
                'EscapeChar': self.escape_char,
                'HeaderLine': self.header_line,
                'DoubleQuote': self.double_quote,
                'QuoteStyle': self.quote_style
            },
            'SCHEMA': {
                'PrimarySchemaConfiguration': self.primary_schema_config
            }
        }

        return self.config_dict

    def config_to_str(self):
        """
        Takes config vars and turns them into a FlexT compatible config string; assigns values to config_str
        :return: returns config_str var
        """
        self.config_str += "[SYNTAX]"
        self.config_str += "\nFileParser=" + self.fileParser
        self.config_str += "\n[CSV]"
        self.config_str += "\nFields=" + ','.join(self.fields)
        self.config_str += "\nDelimiter='" + self.delimiter + "'"
        self.config_str += "\nQuoteChar=" + self.quote_char
        self.config_str += "\nEscapeChar=" + self.escape_char
        self.config_str += "\nHeaderLine=" + str(self.header_line)
        self.config_str += "\nDoubleQuote=" + str(self.double_quote)
        self.config_str += "\nQuoteStyle=" + str(self.quote_style)
        self.config_str += "\n[SCHEMA]"
        self.config_str += "\nPrimarySchemaConfiguration=" + self.primary_schema_config

        return self.config_str



import logging
import os
import csv
from lqmt.lqm.tool import Tool
from lqmt.lqm.data import AlertAction
from lqmt.lqm.data import AlertFields


def replace_empty(value, replace):
    """
    Checks if value is a string and if it is empty. IF true, it's replaced with the provided value. Otherwise, the
    original value is returned
    :param value: Value to check
    :param replace: Value to replace empty string with
    """
    if isinstance(value, str) and not value:
        return replace
    else:
        return value


class ToBro(Tool):
    def __init__(self, config):
        """
        ToFlexText tool. Used to reformat CTI data in a user configured manner.
        :param config: configuration file
        """
        super().__init__(config, [AlertAction.get('All')])
        self._logger = logging.getLogger("LQMT.ToBro.{0}".format(self.getName()))
        self.file = None
        self.writer = None
        self.header_exists = False
        self.header_keys = self.header_formatting()

    def initialize(self):
        super().initialize()

    def process(self, alert):
        """
        Process function. Handles the processing of data for the tool. 
        """
        self.openfile()
        self.start_writer()

        # turn alert data into a csv row and replace empty values with user defined null value (defaults to hyphen)
        row = alert.getFields(self.header_keys)
        row = [replace_empty(c, self._config.null_value) for c in row]

        # write to file
        self.writer.writerow(row)

    def openfile(self):
        """
        Creates and opens the file specified in the user configuration, specifically the file_destination variable.
        """
        if self.file is None:
            try:
                file_dir = os.path.dirname(self._config.file)
                if not os.path.exists(file_dir):
                    os.makedirs(file_dir, 0o755, True)
                if not os.path.exists(self._config.file):
                    open(self._config.file, 'w').close()
                    # create header and open file
                self.bro_header()
                self.file = open(self._config.file, 'a')
            except Exception as e:
                self._logger.error("Unable to open file: {0}".format(self._config.file))
                self._logger.error(e)

    def start_writer(self):
        if self.writer is None and self.file is not None:
            self.writer = csv.writer(self.file, delimiter='\t', quotechar="'")

    def bro_header(self):
        """
        Function that checks file for header. If the file doesn't exist it is created. If the file exists, and there
        is no header, the header is written. Bro's intelligence framework format specifies headers by having the first
        line in the file be '#fields' followed by user defined fields.
        """
        with open(self._config.file, 'r+') as f:
            header = f.readline()
            if "#fields" not in header:
                writer = csv.writer(f, delimiter='\t')
                writer.writerow(["#fields"] + self.header_keys)

    def header_formatting(self):
        if self._config.header_fields is 'all':
            header_keys = list(AlertFields().fields.keys())
            header_keys.sort()
        else:
            header_keys = self._config.header_fields

        return header_keys

    def commit(self):
        pass

    def cleanup(self):
        pass



from lqmt.lqm.tool import Tool
import logging
import pan.xapi
import time
import datetime
from lqmt.lqm.data import AlertAction

class Block:
    """Class to hold information on blocks/revokes"""

    def __init__(self,ip,detected,duration):
        self._ip=ip
        self._detected=int(detected)
        if(duration != None):
            self._duration=int(duration)
        else:
            self._duration=None
    def getAddr(self):
        return self._ip

    def getDetectedTime(self):
        return self._detected

    def setDetectedTime(self, dt):
        self._detected=dt

    def getDuration(self):
        return self._duration

class ToPaloAlto(Tool):
    '''
    '''

    def __init__(self, config):
        super().__init__(config, [AlertAction.get('Block'), AlertAction.get('Revoke')])

        # self._config=config
        self._conn = self._config.getDBConn()
        self._cur = self._conn.cursor()
        self._xapi=self._config.getXapi()
        self._totalProcessed=0
        self._totalBlocked=0
        self._totalRevoked=0
        self._totalUpdated=0
        self._totalPruned=0
        self._totalExpired=0
        self._logger = logging.getLogger("LQMT.PaloAlto.{0}".format(self.getName()))
        self._blocks=dict()
        self._revokes=dict()
        # max per query controls the number of ips per query that are handled.  This is done to ensure that sqlite
        # and python don't have issues if there a significant number of ips to process at once 
        self._maxPerQuery=100

    def initialize(self):
        super().initialize()

    def process(self, alert):
        """Process an alert."""
        self._totalProcessed=self._totalProcessed+1;
        action=alert.getAction()
        if(action == AlertAction.get('Block')):
            self._block(alert)
        elif (action==AlertAction.get('Revoke')):
            self._revoke(alert)

    def commit(self):
        """Commit the changes to the device."""
        # update the database with the new blocks/revokes
        self._updateDB()
        # write the blocks to the EBL files
        self._writeBlocks(self._config.getBlockFiles())
        for ebl in self._config.getBlockLists():
            # tell the device to refresh teach EBL
            try:
                self._xapi.op("request system external-list refresh name \"{0}\"".format(ebl),cmd_xml=True)
            except pan.xapi.PanXapiError as exc:
                self._logger.error("Unable to refresh EBL: {0}".format(ebl))
                self._logger.error(str(exc))

    def cleanup(self):
        self._logger.info("Processed {0} alerts.  New blocks: {1}  Updated Blocks: {2} New revoke: {3} Expired: {4} Pruned: {5}".format(self._totalProcessed,self._totalBlocked,self._totalUpdated,self._totalRevoked,self._totalExpired,self._totalPruned))

    def _updateDB(self):
        """Update the database  - remove any revoked IPs, add any new blocks, and prune if necessary"""
        bk=self._blocks.keys()
        rk=self._revokes.keys()
        allAddrs=set(bk)
        allAddrs.update(rk)
        for addr in allAddrs:
            if(addr in bk and addr in rk):
                # this address appears in both blocks and revokes - use the most recently detected
                b=self._blocks[addr]
                r=self._revokes[addr]
                if(b.getDetectedTime() > r.getDetectedTime()):
                    del self._revokes[addr]
                else:
                    del self._blocks[addr]
        # remove any revoked ips from the database
        if(len(self._revokes) > 0):
            todel=[]
            for r in self._revokes.values():
                todel.append(r.getAddr())
            self._totalRevoked=self._deleteIPsFromDB(todel)
        else:
            self._totalRevoked=0
        ctime=int(time.time())
        blocks=0
        if(len(self._blocks) > 0):
            # if there are blocks,
            # first remove any IPs that already are in the DB (so there are no duplicates)
            # this doesn't currently check to see if an existing block would expire later than a new one 
            # or if the one already in the database was detected before the one about to be added.
            todel=[]
            for b in self._blocks.values():
                todel.append(b.getAddr())
            self._totalUpdated=self._deleteIPsFromDB(todel)

            toadd=list(self._blocks.values())
            # Add the IPs to the DB in _maxPerQuery-sized chunks.
            while(toadd):
                thisadd=toadd[:self._maxPerQuery]
                del toadd[:self._maxPerQuery]
                parms=[]
                for b in thisadd:
                    # if the end time is None, then it is an infinite block and will not be removed until
                    # it is either revoked or removed due to pruning because there are more IPs than the device can handle
                    dur=b.getDuration()
                    if(dur==None):
                        endTime=None
                    else:
                        endTime=ctime+dur
                    parms.append( (b.getAddr(),b.getDetectedTime(),ctime,dur,endTime) )
                self._cur.executemany("insert into blocks (ip,detect_time,start_time,duration,end_time) values (?,?,?,?,?)", parms)
                blocks=blocks+self._cur.rowcount
                self._conn.commit()
        self._totalBlocked=blocks-self._totalUpdated

        # now remove any that have expired
        self._cur.execute("delete from blocks where end_time < ?",(ctime,))
        self._conn.commit()
        self._totalExpired=self._cur.rowcount

        # now that the DB is updated, check to see if it needs to be pruned
        nr=self._getNumRecs()
        toPrune=nr - self._config.getMaxIPsToBlock()
        if(toPrune>0):
            pm=self._config.getPruneMethod()
            # got some pruning to do
            if(pm=='Expiration'):
                self._pruneByExpiration(toPrune,ctime)
            elif (pm=='Added'):
                self._pruneByTimeAdded(toPrune,ctime)
            elif (pm=='Detected'):
                self._pruneByTimeDetected(toPrune,ctime)
        if(toPrune-self._totalPruned > 0):
            #still have more to prune, so use time added as a fall-back
            self._logger.info("Primary pruning method did not remove enough.  Using TimeAdded to prune more")
            tp=self._totalPruned
            self._pruneByTimeAdded(toPrune-self._totalPruned,ctime)
            self._totalPruned=self._totalPruned+tp

    def _deleteIPsFromDB(self,todel):
        """Delete the ips specified in todel from the database."""
        cnt=0
        #self._maxPerQuery limits the number of items in one deletion as a safeguard against a very large query
        while(todel):
            thisdel=todel[:self._maxPerQuery]
            del todel[:self._maxPerQuery]
            self._cur.execute("delete from blocks where ip in ('{0}')".format("','".join(thisdel)))
            self._conn.commit()
            cnt=cnt+self._cur.rowcount
        return cnt

    def _pruneByExpiration(self,toPrune,ctime):
            # delete blocks starting from the soonest to expire until we have reached the limit
            todel=[]
            self._logger.warning("Pruning {0} blocks by expiration:".format(toPrune))
            for r in self._cur.execute("select ip,end_time from blocks where end_time is not null order by end_time limit {0}".format(toPrune)):
                todel.append(r[0])
                expire=datetime.datetime.fromtimestamp(r[1]).strftime("%c")
                self._logger.warning("ip: {0} expiration: {1}: {2} seconds early".format(r[0],expire,r[1]-ctime))
            self._totalPruned=self._deleteIPsFromDB(todel)

    def _pruneByTimeAdded(self,toPrune):
            # delete blocks starting from the least recently added until we have reached the limit
            todel=[]
            for r in self._cur.execute("select ip,start_time,end_time from blocks order by start_time limit {0}".format(toPrune)):
                todel.append(r[0])
                self._logger.warning("Prune: removing block before expiration for ip: {0} expiration: {1}".format(r[0],r[1]))
            self._totalPruned=self._deleteIPsFromDB(todel)

    def _pruneByTimeDetected(self,toPrune):
            # delete blocks starting from the least recently detected until we have reached the limit
            todel=[]
            for r in self._cur.execute("select ip,detect_time,end_time from blocks order by detect_time limit {0}".format(toPrune)):
                todel.append(r[0])
                self._logger.warning("Prune: removing block before expiration for ip: {0} expiration: {1}".format(r[0],r[1]))
            self._totalPruned=self._deleteIPsFromDB(todel)

    def _getNumRecs(self):
        """Return the number of records in the database"""
        self._cur.execute("select count(*) as num from blocks")
        r=self._cur.fetchone()
        return r[0]

    def _revoke(self,alert):
        """Revoke the block of the IP specified in the alert."""
        addr=alert.getIPToBlock()
        if(addr == None):
            # this alert will be unprocessed - put it in the unprocessed file
            self.unprocessed(alert)
            return
        if(self.is_valid_ipv4(addr) or self.is_valid_ipv6(addr)):
            if (addr in self._blocks.keys()):
                r=self._revokes[addr]
                # if tyis is a newer revoke than one already in the list
                if(alert.getDetectedTime() > r.getDetectedTime() ):
                    # update the detected time
                    r.setDetectedTime(alert.getDetectedTime())
            else:
                #otherwise just add it to the revoke list
                self._revokes[addr]=Block(addr,alert.getDetectedTime(),None)
            return True
        return False

    def _block(self,alert):
        """Block the ip from the alert for the duration specified in the alert."""
        addr=alert.getIPToBlock()
        if(addr == None):
            # this alert will be unprocessed - put it in the unprocessed file
            self.unprocessed(alert)
            return
        if(self.is_valid_ipv4(addr) or self.is_valid_ipv6(addr)):
            if (addr in self._blocks.keys()):
                # is this a newer block than we have seen during this round of processing?
                b=self._blocks[addr]
                if(int(alert.getDetectedTime()) > b.getDetectedTime() ):
                    # if so, update the detected time and duration
                    b.setDetectedTime(alert.getDetectedTime())
                    b.setDuration(self._getDuration(alert))
            else:
                #otherwise add it to the blocks list
                self._blocks[addr]=Block(addr,alert.getDetectedTime(),self._getDuration(alert))
            return True
        return False

    def _getDuration(self, alert):
        """Return duration.  If the duration is 0, that means infintite.  If it is None, then use the default duration"""
        duration=alert.getDuration1()
        if(duration==None):
            return self._config.getDefaultDuration()
        if(int(duration)==0):
            return None
        return duration

    def _writeBlocks(self,ipFiles):
        """Write the blocked IPs to the files.  At this point, the database should have no more than can be handled by the device."""
        ipsPerFile=self._config.getIPsPerFile()
        blockFiles=self._config.getBlockFiles()
        curFileIdx=0
        f=None
        n=0
        # retrieve the blocked ips from the database
        for r in self._cur.execute("select ip from blocks order by ip"):
            #strat a new file if necessary
            if(n>=ipsPerFile or f==None):
                if(f!=None):
                    f.close()
                    curFileIdx=curFileIdx+1
                f=open(blockFiles[curFileIdx],"w")
                n=0
            f.write(r[0]+"\n")
            n=n+1
        if(f!=None):
            f.close()
            curFileIdx=curFileIdx+1

        # The palo alto Dynamic Block Lists mechanism doesn't like non-existent or empty EBL files, so make sure there is a "dummy" ip in all that would be empty
        for fn in blockFiles[curFileIdx:]:
            # the PA will have issues with a file with no entries, so write a single VALID ip
            f=open(fn,"w")
            f.write("131.131.131.131\n")
            f.close()

    def fileBegin(self):
        # No special processing when a new file is started
        pass

    def fileDone(self):
        # No special processing when a file is done
        pass



import logging
import os.path
from logging import FileHandler, Formatter
from lqmt.lqm.exceptions import ConfigurationError


class LQMLogging(object):
    """Logging helper class"""

    # by default the debug flag is off
    _debug = False


    @staticmethod
    def isDebug():
        return LQMLogging._debug

    @staticmethod
    def lowerconfig(config):
        """
        Method that makes all keys in the config dict lowercase. Prevents errors where users supply old config style for
        logging.
        :param config: Dict containing configuration data
        :return: Returns config dict with all of it's keys in lowercase.
        """
        newconf = {}
        for key in config.keys():
            newconf[key.lower()] = config[key]

        del config
        return newconf

    def __init__(self, config):
        """Initialize the logging system from the configuration"""

        # TODO: use ezlog to be consistent with other tools in logging
        formatter = Formatter(fmt='%(asctime)s %(name)s %(levelname)s:%(message)s')
        rootLogger = logging.getLogger()
        active = True
        config = self.lowerconfig(config)

        if 'active' in config:
            active = bool(config['active'])

        if active:
            if 'debug' in config:
                LQMLogging._debug = config['debug']
            else:
                LQMLogging._debug = False

            logger = logging.getLogger('LQMT')
            if not LQMLogging._debug:
                logger.setLevel(logging.INFO)
            else:
                logger.setLevel(logging.DEBUG)

            if 'logfilebase' in config:
                lfb = config['logfilebase']

                ldir = os.path.dirname(lfb)
                if (not os.path.exists(ldir)):
                    os.makedirs(ldir, 0o755, True)

                # create handlers/files for error, info, and debug (if necessary)
                errHandler = FileHandler("{0}.err.log".format(lfb))
                errHandler.setLevel(logging.ERROR)
                errHandler.setFormatter(formatter)
                rootLogger.addHandler(errHandler)

                infoHandler = FileHandler("{0}.info.log".format(lfb))
                infoHandler.setLevel(logging.INFO)
                infoHandler.setFormatter(formatter)
                rootLogger.addHandler(infoHandler)

                if LQMLogging._debug:
                    debugHandler = FileHandler("{0}.debug.log".format(lfb))
                    debugHandler.setLevel(logging.DEBUG)
                    debugHandler.setFormatter(formatter)
                    rootLogger.addHandler(debugHandler)

                # The ft handler creates a separate logger for FlexTransform errors
                # There were some files that caused a large number of FlexTransform errors and polluted the regular log
                # files. If it is determined that those issues are gone, then this can be removed
                ftHandler = FileHandler("{0}.ft.log".format(lfb))
                ftHandler.setFormatter(formatter)
                logger = logging.getLogger('FlexTransform')
                logger.addHandler(ftHandler)
                logger.propagate = False
            else:
                logger.error("logfilebase must be specified in configuration")
                raise ConfigurationError("Logfilebase parameter missing")



from lqmt.lqm.tool import ToolConfig


class CEFConfig(ToolConfig):
    def __init__(self, config, csvToolInfo, unhandledCSV):
        super().__init__(config, csvToolInfo, unhandledCSV)
        self._mapping = {}



import re

_valid_varnames = {'hostname', 'port', 'username', 'originator'}

class CheckpointRC():
    def __init__(self, rcfile):
        self.cprc=self._load_file(rcfile)
        
    def _load_file(self,rcfile):
        try:
            f = open(rcfile, 'r')
        except IOError as msg:
            return None

        cprc = {}
        for line in f:
            line = line.rstrip('\r\n')
            if re.search(r'(^#|^\s*$)', line):
                continue
            result = re.search(r'\s*(\w+)\s*=\s*(.+)', line)
            if (result and result.group(1) in _valid_varnames):
                cprc[result.group(1)] = result.group(2)

        f.close()

        return cprc



import logging
import requests
import time
from xml.etree import ElementTree


def create_message(alert):
    """
    Creates a key-value formatted message for Splunk that contains parsed data from FlexT. Defaulting to parsing
    empty values.

    :param alert: Parsed alert data from FlexT
    :return: Returns formatted string.
    """
    data = alert.getAllFields(dictionary=True, parseEmpty=True)
    message = "{0} LQMT: ".format(time.asctime())
    for key, value in data.items():
        message += "{0}={1} ".format(key, value)

    return message


class ApiHandler:
    """
    Class for handling API calls to Splunk's REST Api. This class might end up being redundant depending on a few
    things, but that will be fleshed out further as the tool is built.
    """

    def __init__(self, host=None, port=None, username=None, password=None, splunk_token="", cert_check=True,
                 source=None, sourcetype=None):
        self._messages_processed = 0
        self._logger = logging.getLogger("LQMT.Splunk.ApiCaller")
        self.host = host
        self.port = port
        self.source = source
        self.sourcetype = sourcetype
        self.cert_check = cert_check
        self.url = self.host + ":" + str(self.port)
        self.username = username
        self.password = password
        self.requests = requests
        self.authenticated = False
        self.splunk_token = {'Authorization': splunk_token}
        self.auth_service = "/services/auth/login/"
        self.stream_service = "/services/receivers/stream/"

        # Call authentication function when class object is created.
        self.authenticate()

    def __enter__(self):
        self.authenticate()
        return self

    def __exit__(self):
        self._logger.debug("Total messages processed: {0}".format(self._messages_processed))
        self.requests.post(url=self.url, headers={'Connection': 'close'}, verify=self.cert_check)

    def authenticate(self):
        """
        Method for authenticating against Splunk's REST Api. Returns a session token that will be used for future
        connections. If authentication fails, then the lqmt closes out with an error.
        :return: String: splunk_token. Value containing the splunk token provided by Splunk
        """
        if not self.authenticated:
            data = {'username': self.username, 'password': self.password}
            r = self.requests.post(self.url + self.auth_service, data=data, verify=self.cert_check)

            if r.ok:
                data = ElementTree.fromstring(r.content)
                self.splunk_token['Authorization'] = "Splunk " + data[0].text
                self.authenticated = True
                self._logger.debug(
                    "Successfully authenticated with Splunk instance. Token received: {0}".format(data[0].text)
                )
            else:
                r.raise_for_status()

            return self.splunk_token

    def send_message(self, message, source=None, sourcetype=None):
        """
        Method for sending messages to Splunk via REST api. If the authentication function hasn't been run yet, then
        it is called.
        :param source: Used to override previously set source value
        :param sourcetype: Used to override previously set sourcetype
        :param message: message to be sent to splunk
        """

        if source is not None:
            self.source = source

        if sourcetype is not None:
            self.sourcetype = sourcetype

        if not self.authenticated:
            self.authenticate()

        headers = {"x-splunk-input-mode": "streaming"}
        headers = dict(list(self.splunk_token.items()) + list(headers.items()))
        url = self.url + self.stream_service + "?source={0}&sourcetype={1}".format(self.source, self.sourcetype)
        r = self.requests.post(url, data=message, headers=headers, verify=self.cert_check)
        if r.ok:
            self._messages_processed += 1
        else:
            r.raise_for_status()

        self._logger.debug("Message sent to Splunk. Status code returned: {0}".format(r.status_code))

    def getTotalMessagesProcessed(self):
        return self._messages_processed



"""
Created on Jan 12, 2015

@author: taxon
"""
import FlexTransform
from lqmt.lqm.data import Alert
import os
import inspect
import logging
import io


class FlexTransformParser(object):
    def __init__(self, config=None):
        self._logger = logging.getLogger("LQMT.Parsers")

        # get the FlexTransform directory
        self._current_dir, name = os.path.split(inspect.getfile(FlexTransform.FlexTransform))
        self._transform = FlexTransform.FlexTransform.FlexTransform()

        # config
        if config is not None:
            for p, c in config.items():
                self.addParser(p, c)

    def addParser(self, parserName, parserConfiguration):
        """
        Function that extends the AddParser function from FlexTransform.
        :param parserName: Name of the parser being added
        :param parserConfiguration: Path to the configuration file being added
        """
        if parserName not in self._transform.Parsers:
            config_file = open(os.path.join(self._current_dir, parserConfiguration), 'r')
            self._transform.AddParser(parserName, config_file)

    def parse(self, datafile, meta=None):
        """
        Parse the datafile using the metadata dictionary.

        Note: Removed sourceMetaData from the parser for now due to conflicts with STIX-TLP. Currently assessing
        if it needs to be included. Currently the meta file is just being used to identify the sourceParser.

        :param datafile: Contains path to the file containing the alert data.
        :param meta: Contains meta data about the datafile. Examples includes PayloadFormat, FileName, PayloadType, and
        more.
        :return: Returns parsed alert data.
        """
        alerts = []

        # TODO: Stix-tlp parser currently doesn't support meta files. Until it does, meta files are for the cfm format
        try:
            if meta['PayloadFormat'] == 'stix-tlp' or meta is None:
                data = self._transform.TransformFile(sourceFileName=datafile, sourceParserName=meta['PayloadFormat'],
                                                     targetParserName='LQMTools')
            else:
                data = self._transform.TransformFile(sourceFileName=datafile, sourceParserName=meta['PayloadFormat'],
                                                     targetParserName='LQMTools', sourceMetaData=meta)
        except Exception as e:
            data = []
            self._logger.error(
                "LQMT-FlexTransform-Parser: Error parsing file file='{0}' exception='{1}'".format(datafile, e))

        for d in data:
            alert = Alert()
            alert.setFromDict(d)
            alerts.append(alert)

        return alerts

    def parseflextext(self, datafile, meta, destination_file_obj, config_str):
        """
        Flextext parser. Transforms intermediate data into a user defined structure.

        :param datafile: dir path to the alert datafile that will be transformed
        :param meta: Contains meta data about the datafile. Examples includes PayloadFormat, FileName, PayloadType, and
        more.
        :param destination_file_obj: File object used to write contents to file.
        :param config_str: Parser configuration in the form of a string. Is wrapped in an IO wrapped and then passed
        to FlexT to configure FlexText parser
        :return: Returns none. Data is outputted to file specified by destination_file_obj
        """

        # Add FlexText parser using config_str wrapped in StringIO object
        if 'FlexText' not in self._transform.Parsers:
            config_str_io = io.StringIO(config_str)
            config_str_txt = io.StringIO(config_str_io.read())
            self._transform.AddParser("FlexText", config_str_txt)

        # Run FlexText parser
        try:
            if meta['PayloadFormat'] == 'stix-tlp':
                self._transform.TransformFile(
                    sourceFileName=datafile,
                    targetFileName=destination_file_obj,
                    sourceParserName=meta['PayloadFormat'],
                    targetParserName="FlexText",
                )
            else:
                self._transform.TransformFile(
                    sourceFileName=datafile,
                    targetFileName=destination_file_obj,
                    sourceParserName=meta['PayloadFormat'],
                    targetParserName="FlexText",
                    sourceMetaData=meta
                )

        except Exception as e:
            self._logger.error(
                "LQMT-FlexText-Parser: Error parsing file file='{0}' exception='{1}'".format(datafile, e))


import os
import toml
from unittest import TestCase, main
from lqmt.lqm.tool import ToolConfig
from lqmt.lqm.config import LQMToolConfig
from lqmt.lqm.systemconfig import SystemConfig
from lqmt.test.test_data.sample_config import USERCONFIG


class TestConfiguration(TestCase):
    """
    Testing class for the configuration components
    """

    def setUp(self):
        """
        Initialize the system configuration and the user configuration.

        Note the forward slash replacements for the user configuration. This is due to the forward slash being a
        restricted character in TOML(package used to parse configuration files in LQMT).
        """
        # relative pathing variables. Replace function calls for Windows compatibility.
        self.directory = os.path.dirname(__file__)
        self.alerts = self.directory + "/test_data/"
        self.alerts = self.alerts.replace("\\", "/")
        self.logging = self.directory + "/test_data/test-logs/lqmt"
        self.logging = self.logging.replace("\\", "/")
        self.whitelist = self.directory + "/test_data/whitelist/whitelist.txt"
        self.whitelist = self.whitelist.replace("\\", "/")
        self.whitelist_db = self.directory + "/test_data/whitelist/whitelist.db"
        self.whitelist_db = self.whitelist_db.replace("\\", "/")

        # configurations initialized
        sysconf = SystemConfig()
        self.sys_config = sysconf.getConfig()
        config = USERCONFIG.format(self.alerts, self.logging, self.whitelist, self.whitelist_db)
        self.toml_config = toml.loads(config)
        self.toml_config = self.toml_config["Tools"]["FlexText"][0]  # dirty way of parsing userconfig for ToolConfig
        self.user_config = LQMToolConfig(config)
        self.toolConfig = ToolConfig(self.toml_config, csvToolInfo={""}, unhandledCSV={""})

    def test_user_sources(self):
        """
        Note the replace call. The alert path was replaced with 4 slashes above due to a TOML restriction. After TOML
        parses the config defined above, it strips the extra slashes. We have to do that with our non-stripped alerts
        variable
        """
        sources = self.user_config.getSources().pop()
        self.assertEquals(sources._dirs, [self.alerts.replace("\\\\", "\\")])
        self.assertIsNone(sources.files_to_process)

    def test_user_whitelist(self):
        whitelist = self.user_config.getWhitelist()
        self.assertIsNotNone(whitelist)
        self.assertEqual(whitelist.whitelistFile, self.whitelist.replace("\\\\", "\\"))
        self.assertEqual(whitelist.db, self.whitelist_db.replace("\\\\", "\\"))

    def test_user_parsers(self):
        self.assertIsNotNone(self.user_config.getParser('Cfm13Alert'))
        self.assertIsNotNone(self.user_config.getParser('Cfm20Alert'))
        self.assertIsNotNone(self.user_config.getParser('stix-tlp'))

    def test_user_toollist(self):
        self.assertEqual(self.user_config.getToolsList(), ['FlexText'])

    def test_user_toolchain(self):
        toolchain = self.user_config.getToolChains().pop()
        self.assertEquals(toolchain._name, "anl-flextext-test")
        self.assertEquals(len(toolchain._tools), 1)

    def test_validation(self):
        self.assertTrue(self.toolConfig.validation('incrementFile', bool))
        self.assertEquals(self.toolConfig.validation('fileParser', str, required=True, default="CSV"), 'CSV')
        self.assertEquals(
            self.toolConfig.validation('fields', list, required=True),
            ["indicator", "reportedTime", "detectedTime", "duration1", "priors", "directSource", "reason1", "majorTags",
             "sensitivity", "reconAllowed", "restriction"]
        )


if __name__ == '__main__':
    main()



import os


class ProcessedHandler(object):
    """The ProcessedHandler is the superclass for various implementations
    of what to do with a file once it has been processed"""

    def processed(self, fpath):
        """Called after the file (fpath) has been processed"""
        NotImplemented

    def isProcessed(self, fpath):
        """Called to determine if the file (fpath) has been processed"""
        NotImplemented

    def enteringDirectory(self, path):
        """Called when the directory (fpath) is about to be entered by the DirectorySource"""
        pass

    def leavingDirectory(self, path):
        """Called when the directory (fpath) is about to be left by the DirectorySource"""
        pass

    def skipDirectory(self, path):
        """Called to determine if the file (fpath) should be skipped.
        Return True if the directory should be skipped"""
        pass


class ProcessHandlerMove(ProcessedHandler):
    """This class implements moving the processed file to another directory after processing"""

    def __init__(self):
        # _pdir is the directory that will be created as a child directory in the directory that contains the processed file.
        self._pdir = "processed"
        self._curdir = None

    def processed(self, fpath):
        # move the processed file to _pdir
        fdir, fn = os.path.split(fpath)
        os.rename(os.path.join(fdir, fn), os.path.join(self._curdir, fn))
        os.rename(os.path.join(fdir, "." + fn), os.path.join(self._curdir, "." + fn))

    def isProcessed(self, fpath):
        return False

    def enteringDirectory(self, path):
        # if the _pdir doesn't ewxist for this path, create it
        self._curdir = os.path.join(path, self._pdir)
        if (not os.path.exists(self._curdir)):
            os.mkdir(self._curdir)

    def skipDirectory(self, path):
        # skip this directory if it is the _pdir directory
        return os.path.basename(path) == self._pdir


class ProcessHandlerDelete(ProcessedHandler):
    """This ProcessedHandler class implements deleting the processed file after processing"""

    def processed(self, fpath):
        # delete the file
        fdir, fn = os.path.split(fpath)
        os.unlink(os.path.join(fdir, fn))
        os.unlink(os.path.join(fdir, "." + fn))

    def isProcessed(self, fpath):
        return False


class ProcessHandlerDoNothing(ProcessedHandler):
    """This ProcessedHandler class implements doing nothing after processing.
    This should only be used for debugging"""

    def processed(self, fpath):
        pass

    def isProcessed(self, fpath):
        return False


class ProcessHandlerTrackFile(ProcessedHandler):
    """This ProcessedHandler class implements adding the file's name to a list of files processed that is stored in the
    directory. This is not a terribly efficient method, as the file fill grow and so will the directory size"""

    def __init__(self, trackFile):
        self._processed = None
        self._pfile = None
        self._trackFile = trackFile

    def processed(self, fpath):
        self._pfile.write(os.path.basename(fpath) + '\n')
        self._pfile.flush()
        os.fsync(self._pfile.fileno())

    def isProcessed(self, fpath):
        return os.path.basename(fpath) in self._processed

    def enteringDirectory(self, path):
        pfile = os.path.join(path, self._trackFile)
        try:
            self._processed = {}
            open(pfile, 'a').close()  # ensures file is writable
            f = open(pfile, 'r')
            for line in f:
                self._processed[line.rstrip()] = 1
            f.close()
            self._pfile = open(pfile, 'a')
        except IOError as e:
            raise ValueError('procfile (%s) generated an IO Error: %s\n' % (pfile, e.strerror))

    def leavingDirectory(self, path):
        self._pfile.close()
        self._processed = None



import logging
import socket
from lqmt.lqm.data import AlertAction
from lqmt.lqm.logging import LQMLogging
from lqmt.lqm.tool import Tool


class ToSysLog(Tool):
    """ToSysLog communicates with the SysLog logger via udp or tcp based on the user config settings."""

    def __init__(self, config):
        super().__init__(config, [AlertAction.get('All')])
        self._logger = logging.getLogger("LQMT.SysLog.{0}".format(self.getName()))
        self._totalSent = 0
        self._messageHead = self._config.messageHead
        self._messageFields = self._config.messageFields

        if self.isEnabled():
            if self._config.protocol == 'tcp':
                self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                try:
                    self._socket.connect((self._config.host, self._config.port))
                except Exception as e:
                    self._logger.error(
                            "Unable to connect to remote Syslog server: {0}:{1}".format(self._config.host,
                                                                                        self._config.port))
                    self._logger.error(str(e))
                    self._socket = None
                    self.disable()
            else:
                self._socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

    def initialize(self):
        super().initialize()

    def process(self, data):
        """
        Process alerts to SysLog server. Message format defined in user config file

        :param data: alert data
        """

        if not self.isEnabled():
            return

        try:
            # format data being sent to syslog
            msg = self.format_syslog_msg(data.getFields(self._messageFields))
            syslog_msg = bytes("{0}\n".format(msg), 'UTF-8')
            if self._config.protocol == "TCP":
                self._socket.send(syslog_msg)
            else:
                self._socket.sendto(syslog_msg, (self._config.host, self._config.port))
            self._totalSent += 1
        except Exception as e:
            msg = "Error while sending data to remote Syslog server"
            if not LQMLogging.isDebug():
                self._logger.error(msg)
            else:
                self._logger.exception(msg)
            self._logger.error(str(e))

    def commit(self):
        pass

    def cleanup(self):
        self._logger.info("Send {0} messages to Syslog server".format(self._totalSent))
        if self.isEnabled():
            try:
                self._socket.close()
            except Exception as e:
                self._logger.error("Error while closing connection to remote Syslog server")
                self._logger.error(str(e))

    def format_syslog_msg(self, extractedfields):
        msg = self._messageHead
        step = 0  # used to iterate through extracted fields during format process

        # begin formatting...
        for i in self._messageFields:
            # extract value from extractedFields
            value = extractedfields[step]

            # check for existing quotes and strip
            if value.startswith('"') and value.endswith('"'):
                value = value[1:-1]
            # append value to message
            msg += i + '="%s" ' % value
            step += 1
        return msg



'''
Created on Oct 23, 2014

@author: taxon
'''
import socket
from lqmt.lqm.tool import Tool
import logging
from lqmt.lqm.data import AlertAction
from lqmt.lqm.logging import LQMLogging


class ToArcSight(Tool):
    """ToArcSight communicates with the ArcSight logger via udp or tcp based on the configuration settings"""

    def __init__(self, config):
        super().__init__(config, [AlertAction.get('All')])
        self._logger = logging.getLogger("LQMT.ArcSight.{0}".format(self.getName()))
        self._totalSent = 0

        if self.isEnabled():
            # Open the appropriate type of socket
            if self._config.getProtocol() == 'tcp':
                self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                try:
                    self._socket.connect((config.getHost(), config.getPort()))
                except Exception as e:
                    self._logger.error(
                        "Unable to connect to ArcSight: {0}:{1}".format(config.getHost(), config.getPort()))
                    self._logger.error(str(e))
                    self._socket = None
                    self.disable()
            else:
                self._socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

    def initialize(self):
        super().initialize()

    def process(self, data):
        """Pass the alert to the ArcSight Logger instance.  data is a CEF string."""
        if not self.isEnabled():
            return
        s = data.toCEFString()
        try:
            msg = bytes("{0}\n".format(s), 'UTF-8')
            if self._config.getProtocol() == 'tcp':
                self._socket.send(msg)
            else:
                self._socket.sendto(msg, (self._config.getHost(), self._config.getPort()))
            self._totalSent += 1
        except Exception as e:
            msg = "Error while sending data to ArcSight server"
            if not LQMLogging.isDebug():
                self._logger.error(msg)
            else:
                self._logger.exception(msg)
            self._logger.error(str(e))

    def commit(self):
        pass

    def cleanup(self):
        self._logger.info("Sent {0} messages to ArcSight server".format(self._totalSent))
        if self.isEnabled():
            try:
                self._socket.close()
            except Exception as e:
                self._logger.error("Error while closing connection to ArcSight server")
                self._logger.error(str(e))



import logging
from lqmt.lqm.tool import ToolConfig
from lqmt.lqm.exceptions import ConfigurationError


class ArcSightConfig(ToolConfig):
    def __init__(self, configData, csvToolInfo, unhandledCSV):
        super().__init__(configData, csvToolInfo, unhandledCSV)
        self.logger = logging.getLogger("LQMT.ArcSight.{0}".format(self.getName()))
        hasError = False

        if 'host' in configData:
            self._host = configData['host']
        else:
            self._logger.error("host must be specified in the configuration")
            hasError = True
        if 'port' in configData:
            self._port = configData['port']
        else:
            self._logger.error("port must be specified in the configuration")
            hasError = True
        if 'protocol' in configData:
            protocol = configData['protocol']
            if protocol == 'udp' or protocol == 'tcp':
                self._protocol = protocol
            else:
                self._logger.error("Invalid protocol: {0}".format(protocol))
                hasError = True
            self._protocol = configData['protocol']
        else:
            self._logger.error("protocol must be specified in the configuration")
            hasError = True

        if hasError:
            self.disable()
            raise ConfigurationError("Missing a required value in the user configuration for the to_arcsight tool")

    def getHost(self):
        return self._host

    def getPort(self):
        return self._port

    def getProtocol(self):
        return self._protocol



import os
from unittest import TestCase, main
import lqmt.whitelist.master as master


class TestWhitelist(TestCase):
    """
    Testing class for whitelist function. Used to test if a value belongs in the user defined whitelist or not.
    """

    def setUp(self):
        currentdir = os.path.dirname(__file__)
        whitelist = currentdir + "/test_data/whitelist/whitelist.txt"
        db = currentdir + "/test_data/whitelist/whitelist.db"
        configstr = """
        [Whitelist]
            whitelist = '{0}'
            dbfile = '{1}'
        """.format(whitelist, db)

        self.wl = master.MasterWhitelist(configStr=configstr)
        self.indicatorTypes = master.IndicatorTypes

    # IPV4 Addresses
    def test_match_by_ipv4(self):
        self.assertTrue(self.wl.isWhitelisted(self.indicatorTypes.ipv4, '192.168.1.1'))

    def test_failed_match_by_ipv4(self):
        self.assertFalse(self.wl.isWhitelisted(self.indicatorTypes.ipv4, '192.168.1.2'))

    def test_match_ipv4_by_subnet(self):
        self.assertTrue(self.wl.isWhitelisted(self.indicatorTypes.ipv4, '192.168.2.5'))

    # IPV4 Subnets
    def test_match_ipv4_sub_by_inclusion(self):
        self.assertTrue(self.wl.isWhitelisted(self.indicatorTypes.ipv4subnet, '192.168.2.0/26'))

    def test_match_contiguous_ipv4_sub(self):
        self.assertTrue(self.wl.isWhitelisted(self.indicatorTypes.ipv4subnet, '192.168.2.0/23'))

    def test_failed_match_large_ipv4_sub(self):
        self.assertFalse(self.wl.isWhitelisted(self.indicatorTypes.ipv4subnet, '192.168.2.0/22'))

    def test_failed_match_ipv4_sub(self):
        self.assertFalse(self.wl.isWhitelisted(self.indicatorTypes.ipv4subnet, '192.168.20/24'))

    # IPV6 Addresses
    def test_match_by_ipv6(self):
        self.assertTrue(self.wl.isWhitelisted(self.indicatorTypes.ipv6, '6EE2:317F:0AF1:684C:AAC8:43F4:6E49:7D86'))

    def test_failed_match_by_ipv6(self):
        self.assertFalse(self.wl.isWhitelisted(self.indicatorTypes.ipv6, '6EE2:317F:0AF1:684C:AAC8:43F4:6E49:AAAA'))

    # Domains
    def test_match_domain(self):
        self.assertTrue(self.wl.isWhitelisted(self.indicatorTypes.domain, 'thegoodguys.com'))

    def test_failed_match_domain(self):
        self.assertFalse(self.wl.isWhitelisted(self.indicatorTypes.domain, 'badguy.com'))

    # Host
    def test_match_host_by_domain(self):
        self.assertTrue(self.wl.isWhitelisted(self.indicatorTypes.host, 'oneof.thegoodguys.com'))

    def test_match_host(self):
        self.assertTrue(self.wl.isWhitelisted(self.indicatorTypes.host, 'really.goodguy.com'))

    def test_failed_match_host_by_url(self):
        self.assertFalse(self.wl.isWhitelisted(self.indicatorTypes.url, 'very.badguy.com'))

    # URL
    def test_match_by_url(self):
        self.assertTrue(
            self.wl.isWhitelisted(self.indicatorTypes.url, 'http://www.goodsite.com/blah?someparm=9&parm2=foo#anchor'))

    def test_failed_match_by_url(self):
        self.assertFalse(
            self.wl.isWhitelisted(self.indicatorTypes.url, 'http://www.badsite.com/blah?someparm=9&parm2=foo#anchor'))


if __name__ == '__main__':
    main()



from lqmt.whitelist.master import IndicatorTypes


class AlertAction(object):
    """Enum of all possible actions"""
    enums = ["All", "Block", "Notify", "Watch", "SendReport", "Revoke", "OtherAction"]

    @staticmethod
    def get(action):
        if action in AlertAction.enums:
            return action
        raise Exception("Invalid action specified: {0}".format(action))


class AlertFields(object):
    """Class to aid in properly formatting fields as strings and checking for existence of alert fields"""

    def __init__(self):
        self._fields = {"dataItemID": "S", "fileID": "S", "detectedTime": "I", "reportedTime": "I",
                        "processedTime": "I", "indicator": "S", "indicatorType": "S", "indicatorDirection": "S",
                        "secondaryIndicator": "S", "secondaryIndicatorType": "S", "secondaryIndicatorDirection": "S",
                        "directSource": "S", "secondarySource": "S", "action1": "S", "duration1": "I", "action2": "S",
                        "duration2": "I", "reason1": "S", "reference1": "S", "reason2": "S", "reference2": "S",
                        "majorTags": "S", "minorTags": "S", "restriction": "S", "sensitivity": "S", "reconAllowed": "S",
                        "priors": "I", "confidence": "I", "severity": "I", "relevancy": "I", "relatedID": "S",
                        "relationType": "S", "comment": "S", "fileHasMore": "I"
                        }

    def isValid(self, field):
        return field in self._fields

    def getStringRepresentation(self, field, val):
        """Return a string representation of the specified field"""
        if not self.isValid(field):
            raise Exception("Alert field {0} is not a valid field".format(field))
        ftype = self._fields[field]
        if ftype == "I":
            if val is not None:
                return val
            else:
                return ""
        elif ftype == "S":
            if val is not None:
                return '"' + val + '"'
            else:
                return ""
        elif ftype == "E":
            if val is not None:
                return '"' + val.name + '"'
            else:
                return ""

    @property
    def fields(self):
        return self._fields


class Alert(object):
    """The Alert object represents the LQM intermediate format"""
    _alertFields = AlertFields()

    @staticmethod
    def isValidField(field):
        return Alert._alertFields.isValid(field)

    def __init__(self):
        self._dataItemID = None
        self._fileID = None
        self._detectedTime = None
        self._reportedTime = None
        self._processedTime = None
        self._indicator = None
        self._indicatorType = None
        self._indicatorDirection = None
        self._secondaryIndicator = None
        self._secondaryIndicatorType = None
        self._secondaryIndicatorDirection = None
        self._directSource = None
        self._secondarySource = None
        self._action1 = None
        self._duration1 = None
        self._action2 = None
        self._duration2 = None
        self._reason1 = None
        self._reference1 = None
        self._reason2 = None
        self._reference2 = None
        self._majorTags = None
        self._minorTags = None
        self._restriction = None
        self._sensitivity = None
        self._reconAllowed = None
        self._priors = None
        self._confidence = None
        self._severity = None
        self._relevancy = None
        self._relatedID = None
        self._relationType = None
        self._comment = None
        self._fileHasMore = None

    # setters
    def setFromDict(self, d):
        """Set the fields from the dictionary"""
        if 'dataItemID' in d:
            self._dataItemID = d['dataItemID']
        if 'fileID' in d:
            self._fileID = d['fileID']
        if 'detectedTime' in d:
            self._detectedTime = d['detectedTime']
        if 'reportedTime' in d:
            self._reportedTime = d['reportedTime']
        if 'processedTime' in d:
            self._processedTime = d['processedTime']
        if 'indicator' in d:
            self._indicator = d['indicator']
        if 'indicatorType' in d:
            self._indicatorType = d['indicatorType']
            if self._indicatorType == "URL":
                pass
        if 'indicatorDirection' in d:
            self._indicatorDirection = d['indicatorDirection']
        if 'secondaryIndicator' in d:
            self._secondaryIndicator = d['secondaryIndicator']
        if 'secondaryIndicatorType' in d:
            self._secondaryIndicatorType = d['secondaryIndicatorType']
        if 'secondaryIndicatorDirection' in d:
            self._secondaryIndicatorDirection = d['secondaryIndicatorDirection']
        if 'directSource' in d:
            self._directSource = d['directSource']
        if 'secondarySource' in d:
            self._secondarySource = d['secondarySource']
        if 'action1' in d:
            self._action1 = AlertAction.get(d['action1'])
        if 'duration1' in d:
            self._duration1 = d['duration1']
        if 'action2' in d:
            self._action2 = AlertAction.get(d['action2'])
        if 'duration2' in d:
            self._duration2 = d['duration2']
        if 'reason1' in d:
            self._reason1 = d['reason1']
        if 'reference1' in d:
            self._reference1 = d['reference1']
        if 'reason2' in d:
            self._reason2 = d['reason2']
        if 'reference2' in d:
            self._reference2 = d['reference2']
        if 'majorTags' in d:
            self._majorTags = d['majorTags']
        if 'minorTags' in d:
            self._minorTags = d['minorTags']
        if 'restriction' in d:
            self._restriction = d['restriction']
        if 'sensitivity' in d:
            self._sensitivity = d['sensitivity']
        if 'reconAllowed' in d:
            self._reconAllowed = d['reconAllowed']
        if 'priors' in d:
            self._priors = d['priors']
        if 'confidence' in d:
            self._confidence = d['confidence']
        if 'severity' in d:
            self._severity = d['severity']
        if 'relevancy' in d:
            self._relevancy = d['relevancy']
        if 'relatedID' in d:
            self._relatedID = d['relatedID']
        if 'relationType' in d:
            self._relationType = d['relationType']
        if 'comment' in d:
            self._comment = d['comment']
        if 'fileHasMore' in d:
            self._fileHasMore = d['fileHasMore']

    def isWhitelisted(self, wl):
        """Return whether or not this Alert is whitelisted"""
        if wl is None:
            return False
        # check both primary and secondary indicators (if they exist)
        if self._indicator is not None:
            if wl.isWhitelisted(self._getIndicatorType(self._indicatorType), self._indicator):
                return True
        if self._secondaryIndicator is not None:
            if wl.isWhitelisted(self._getIndicatorType(self._secondaryIndicatorType), self._secondaryIndicator):
                return True
        return False

    def _getIndicatorType(self, indType):
        """Return the enumerated indicator type of the indType"""
        if indType == "IPv4Address":
            return IndicatorTypes.ipv4
        elif indType == "IPv6Address":
            return IndicatorTypes.ipv6
        elif indType == "DNSDomainName":
            return IndicatorTypes.domain
        elif indType == "DNSHostName":
            return IndicatorTypes.host
        elif indType == "URL":
            return IndicatorTypes.url

    def setDataItemID(self, dataItemID):
        self._dataItemID = dataItemID

    def setFileID(self, fileID):
        self._fileID = fileID

    def setDetectedTime(self, detectedTime):
        self._detectedTime = detectedTime

    def setReportedTime(self, reportedTime):
        self._reportedTime = reportedTime

    def setProcessedTime(self, processedTime):
        self._processedTime = processedTime

    def setIndicator(self, indicator):
        self._indicator = indicator

    def setIndicatorType(self, indicatorType):
        self._indicatorType = indicatorType

    def setIndicatorDirection(self, indicatorDirection):
        self._indicatorDirection = indicatorDirection

    def setSecondaryIndicator(self, secondaryIndicator):
        self._secondaryIndicator = secondaryIndicator

    def setSecondaryIndicatorType(self, secondaryIndicatorType):
        self._secondaryIndicatorType = secondaryIndicatorType

    def setSecondaryIndicatorDirection(self, secondaryIndicatorDirection):
        self._secondaryIndicatorDirection = secondaryIndicatorDirection

    def setDirectSource(self, directSource):
        self._directSource = directSource

    def setSecondarySource(self, secondarySource):
        self._secondarySource = secondarySource

    def setAction1(self, action1):
        self._action1 = action1

    def setDuration1(self, duration1):
        self._duration1 = duration1

    def setAction2(self, action2):
        self._action2 = action2

    def setDuration2(self, duration2):
        self._duration2 = duration2

    def setReason1(self, reason1):
        self._reason1 = reason1

    def setReference1(self, reference1):
        self._reference1 = reference1

    def setReason2(self, reason2):
        self._reason2 = reason2

    def setReference2(self, reference2):
        self._reference2 = reference2

    def setMajorTags(self, majorTags):
        self._majorTags = majorTags

    def setMinorTags(self, minorTags):
        self._minorTags = minorTags

    def setRestriction(self, restriction):
        self._restriction = restriction

    def setSensitivity(self, sensitivity):
        self._sensitivity = sensitivity

    def setReconAllowed(self, reconAllowed):
        self._reconAllowed = reconAllowed

    def setPriors(self, priors):
        self._priors = priors

    def setConfidence(self, confidence):
        self._confidence = confidence

    def setSeverity(self, severity):
        self._severity = severity

    def setRelevancy(self, relevancy):
        self._relevancy = relevancy

    def setRelatedID(self, relatedID):
        self._relatedID = relatedID

    def setRelationType(self, relationType):
        self._relationType = relationType

    def setComment(self, comment):
        self._comment = comment

    def setFileHasMore(self, fileHasMore):
        self._fileHasMore = fileHasMore

    # getters
    def getDataItemID(self):
        return self._dataItemID

    def getDetectedTime(self):
        return self._detectedTime

    def getProcessedTime(self):
        return self._processedTime

    def getDuration1(self):
        return self._duration1

    def getIPToBlock(self):
        if self.getAction() == AlertAction.get('Block') and self._indicatorType == "IPv4Address":
            return self._indicator
        else:
            return None

    def getIPToRevoke(self):
        if self.getAction() == AlertAction.get('Revoke') and self._indicatorType != "IPv4Address":
            return self._indicator
        else:
            return None

    def getSourceIP(self):
        if self._indicatorDirection is None or self._indicatorDirection == "destination":
            return None
        if self._indicatorType is None:
            return None
        if self._indicatorType != "IPv4Address":
            return None
        return self._indicator

    def getDestIP(self):
        if self._indicatorDirection is None or self._indicatorDirection == "source":
            return None
        if self._indicatorType is None:
            return None
        if self._indicatorType != "IPv4Address":
            return None
        return self._indicator

    def getURL(self):
        if self._indicatorType != "URL":
            return None
        return self._indicator

    def getSourceHost(self):
        if self._indicatorDirection is None or self._indicatorDirection == "destination":
            return None
        if self._indicatorType is None:
            return None
        if self._indicatorType != "DNSHostName":
            return None
        return self._indicator

    def getDestHost(self):
        if self._indicatorDirection is None or self._indicatorDirection == "source":
            return None
        if self._indicatorType is None:
            return None
        if self._indicatorType != "DNSHostName":
            return None
        return self._indicator

    def getSourcePort(self):
        # parser currently doesn't support secondaryIndicators
        if self._secondaryIndicatorDirection is None or self._secondaryIndicatorDirection == "destination":
            return None
        if self._secondaryIndicatorType is None:
            return None
        if self._secondaryIndicatorType != "tcpport" or self._secondaryIndicatorType != "udpport":
            return None
        return self._secondaryIndicator

    def getDestPort(self):
        # parser currently doesn't support secondaryIndicators
        if self._secondaryIndicatorDirection is None or self._secondaryIndicatorDirection == "source":
            return None
        if self._secondaryIndicatorType is None:
            return None
        if self._secondaryIndicatorType != "tcpport" or self._secondaryIndicatorType != "udpport":
            return None
        return self._secondaryIndicator

    def getTransportProtocol(self):
        # parser currently doesn't support secondaryIndicators
        if self._secondaryIndicatorDirection is None or self._secondaryIndicatorDirection == "destination":
            return None
        if self._secondaryIndicatorType is None:
            return None
        if self._secondaryIndicatorType == "tcpport":
            return "TCP"
        else:
            return "UDP"

    def getDirectSource(self):
        return self._directSource

    def getAction(self):
        return self._action1

    def getEndTime(self):
        # this could be processedTime + duration or reportedTime + duration or detectedTime + duration
        #  for now use none of those
        return None

    def getReason(self):
        return self._reason1

    def getPriors(self):
        return self._priors

    def getComment(self):
        return self._comment

    def getRestriction(self):
        return self._restriction

    def getSensitivity(self):
        return self._sensitivity

    def getReconAllowed(self):
        return self._reconAllowed

    def getFileID(self):
        return self._fileID

    def getReportedTime(self):
        return self._reportedTime

    def getSecondarySource(self):
        return self._secondarySource

    def getConfidence(self):
        return self._confidence

    def getSeverity(self):
        return self._severity

    def getRelevance(self):
        return self._relevancy

    def getRelatedID(self):
        return self._relatedID

    def getRelationType(self):
        return self._relationType

    def getFileHasMore(self):
        return self._fileHasMore

    def getFields(self, fieldNames):
        fields = []
        for field in fieldNames:
            fields.append(Alert._alertFields.getStringRepresentation(field, self._getField(field)))
        return fields

    def _getField(self, field):
        val = None
        tfld = "_" + field
        if hasattr(self, tfld):
            val = getattr(self, tfld)
        return val

    def getAllFields(self, dictionary=False, parseEmpty=False, emptyValue=None):
        """
        Method to get all supported fields from the intermediate data format.
        :param dictionary: Option to return the results as a dictionary. Defaults to False.
        :param parseEmpty: Option to parse out any empty values.
        :param emptyValue: Option to fill in empty fields with a different value
        :return: Returns either a list or dictionary of all fields and their parsed value. Defaults to a list.
        """
        keys = list(Alert._alertFields.fields.keys())
        fields = self.getFields(keys)
        if dictionary:
            dict_fields = {}
            keys.reverse()
            for value in fields:
                if value is not "":
                    dict_fields[keys.pop()] = value
                elif parseEmpty:
                    keys.pop()
                elif emptyValue:
                    dict_fields[keys.pop()] = '"{0}"'.format(emptyValue)
            return dict_fields
        else:
            return fields



import logging
from lqmt.lqm.tool import ToolConfig


class SplunkConfig(ToolConfig):

    def __init__(self, configData, csvToolInfo, unhandledCSV):
        super().__init__(configData, csvToolInfo, unhandledCSV)
        self.logger = logging.getLogger("LQMT.Splunk.{0}".format(self.getName()))

        self.host = self.validation('host', str, required=True)
        self.port = self.validation('port', int, default=8089)
        self.username = self.validation('username', str, required=True)
        self.password = self.validation('password', str, required=True)
        self.cert_check = self.validation('cert_check', bool, default=True)
        self.source = self.validation('source', str, default="lqmt")
        self.sourcetype = self.validation('sourcetype', str, required=True)



from lqmt.lqm.tool import ToolConfig
from lqmt.lqm.config import ConfigurationError
import logging
import pan.xapi
import os.path
import sqlite3
from lxml import objectify
import ssl
import sys


class PaloAltoConfig(ToolConfig):
    def __init__(self, config, csvToolInfo, unhandledCSV):
        super().__init__(config, csvToolInfo, unhandledCSV)
        self._logger = logging.getLogger("LQMT.PaloAlto.{0}".format(self.getName()))
        hasError = False

        if ('api_key' in config):
            self._apiKey = config['api_key']
        else:
            self._apiKey = None
        if ('api_username' in config):
            self._userName = config['api_username']
        else:
            self._userName = None
        cur_version = sys.version_info
        if ('cafile' in config):
            if (cur_version < (3, 4)):
                self._logger.info("cafile specified, but not needed due to python version")
            self._cafile = config['cafile']
        else:
            if (cur_version >= (3, 4)):
                self._logger.error("cafile must be specified in the configuration")
                hasError = True
            self._cafile = None
        if ('api_password' in config):
            self._password = config['api_password']
        else:
            self._password = None
        if ('hostname' in config):
            self._hostname = config['hostname']
        else:
            self._logger.error("hostname must be specified in the configuration")
            hasError = True
        if ('badIPFiles' in config):
            self._badIPFiles = config['badIPFiles']
        else:
            self._logger.error("badIPFiles must be specified in the configuration")
            hasError = True
        if ('default_duration' in config):
            self._defaultDuration = config['default_duration']
        else:
            self._defaultDuration = 3 * 24 * 3600
            self._logger.warning(
                "default_duration not specified in the configuration, setting default_duration to 86400 seconds")
        if ('block_lists' in config):
            self._block_lists = config['block_lists']
        else:
            self._logger.error("block_lists must be specified in the configuration")
            hasError = True
        if (len(self._badIPFiles) != len(self._block_lists)):
            self._logger.error("The number of block list names and bad IP file MUST be the same")
            hasError = True

        if ('db_location' in config):
            self._db_location = config['db_location']
        else:
            self._db_location = "pa-{0}".format(self.getName())
            self._logger.info("db_location not specified using:  must be specified in the configuration")
            hasError = True
        if ('prune_method' in config):
            pm = config['prune_method']
            valid_emthods = ['Expiration', 'Added', 'Detected']
            if (pm in valid_emthods):
                self._prune_method = pm
            else:
                self._logger.error("Invalid prune method: {0}".format(pm))
        else:
            self._prune_method = 'Expiration'
        if (self._apiKey == None):
            if (self._userName == None or self._password == None):
                self._logger.error("Either the api_key or username/password must be specified in the configuration")
                hasError = True
        else:
            if (self._userName != None or self._password != None):
                self._logger.error(
                    "Only one of the api_key or username/password must be specified in the configuration")
                hasError = True
        if (hasError):
            self.disable()
            raise ConfigurationError("Missing a required value in the user configuration for the to_paloalto tool")
        ctx = None
        cur_version = sys.version_info
        if (cur_version >= (3, 4)):
            # required by pan api if python 3.4 is used
            ctx = ssl.create_default_context(cafile=self._cafile)
        self._xapi = pan.xapi.PanXapi(api_username=self._userName, api_password=self._password, api_key=self._apiKey,
                                      hostname=self._hostname, ssl_context=ctx)
        self._ipsPerBL = self._getIPsPerBL()
        self._ensureDBExists()

    def getDBConn(self):
        return self._conn

    def _ensureDBExists(self):
        dbfile = os.path.join(self._db_location, "{0}.db".format(self.getName()))
        if (not os.path.exists(self._db_location)):
            os.makedirs(self._db_location, 0o755, True)
        if (not os.path.exists(dbfile)):
            conn = sqlite3.connect(dbfile)
            c = conn.cursor()
            c.execute(
                "create table blocks (ip text primary key, detect_time integer(11), start_time integer(11), end_time integer(11), duration integer(11))")
            c.execute("create index detect_time on blocks(start_time)")
            c.execute("create index start_time on blocks(start_time)")
            c.execute("create index end_time on blocks(end_time)")
            conn.close()
        self._conn = sqlite3.connect(dbfile)

    def _getIPsPerBL(self):
        #        self._xapi.op("<show><system><state><filter>cfg.general.max-address</filter></state></system></show>")
        try:
            self._xapi.op("show system state filter \"cfg.general.max-address\"", cmd_xml=True)
            d = self._xapi.xml_root()
            o = objectify.fromstring(d)
            o.result
            s = o.result.text.split(" ")[1].strip();
            if (s.startswith("0x")):
                return int(s, 16) - 300
            else:
                return int(s) - 300
        except Exception as e:
            self._logger.error("Unable to retrieve max ip addresses")
            self._logger.error(str(e))
            self.disable()
            return 0

    def getXapi(self):
        return self._xapi

    def getBlockLists(self):
        return self._block_lists

    def getBlockFiles(self):
        return self._badIPFiles

    def getDefaultDuration(self):
        return self._defaultDuration

    def getMaxIPsToBlock(self):
        return self._ipsPerBL * len(self._badIPFiles)

    def getIPsPerFile(self):
        return self._ipsPerBL

    def getPruneMethod(self):
        return self._prune_method



from . import processed
import os
from .sources import Source
from lqmt.lqm.exceptions import ConfigurationError
import logging


class FilesToProcess(object):
    """Implements a directory traversal of each of the top-level dirs this object is initialized with"""

    def __init__(self, dirs, postProcess):
        self._iters = []
        self._dirs = dirs
        self._dirIter = iter(self._dirs)
        self._curFiles = None
        self._curDir = None
        self._postProcess = postProcess
        self._numFiles = 0
        self._numDirs = 0

        self._getNextTLD()

    def _getNextTLD(self):
        # get the next top-level directory
        self._currentTLD = next(self._dirIter)
        # and put the iterator for the list of files in the directory on the _iters list
        self._iters.append((self._currentTLD, iter(os.listdir(self._currentTLD))))

    def _advanceToNextFile(self):
        """Advance to the next file"""
        found = False
        while not found:
            # while a valid file pair (file/metadata file) has not been found
            # get the last dir & iter on the iters list
            dirName, cDirIter = self._iters[-1]
            try:
                # get the next entry from the dir
                entry = next(cDirIter)
                path = dirName + "/" + entry
                if os.path.isdir(path):
                    # if it is a path and it is not to be skipped,
                    # append an iterator of the directory's contents
                    if not self._postProcess.skipDirectory(path):
                        self._iters.append((path, iter(os.listdir(path))))
                elif not entry.startswith("."):
                    # otherwise, if the entry doesn't start with a '.'
                    if os.path.isfile(path):
                        # and it is a file, check to see if a metadata file exists
                        if os.path.exists(dirName + "/." + entry):
                            # if there is a matching metadata file
                            # check to see if we have left a directory
                            if self._curDir != dirName:
                                # if we have, then tell the post processor
                                if self._curDir is not None:
                                    self._postProcess.leavingDirectory(self._curDir)
                                self._curDir = dirName
                                self._postProcess.enteringDirectory(self._curDir)
                                self._numDirs += 1
                            self._curDir = dirName
                            # if the file hasn't already been processed, then we found the next file
                            if not self._postProcess.isProcessed(path):
                                # so set the flag to exit the loop and save the file info for retrieval
                                found = True
                                self._curFiles = (path, dirName + "/." + entry)
            except StopIteration:
                # if the iteration of the current dir is done
                # get the next dir, if any, and continue
                self._iters.pop()
                if len(self._iters) == 0:
                    self._getNextTLD()
        if not found:
            if self._curDir is not None:
                self._postProcess.leavingDirectory(self._curDir)
            raise StopIteration()

    def getNextFile(self):
        self._advanceToNextFile()
        self._numFiles += 1
        return self._curFiles

    def __iter__(self):
        return self

    def __next__(self):
        return self.getNextFile()


class DirectorySource(Source):
    """ This source provides an iterator for all the file pairs (alert/metadata) contained in all of
    its children that haven't already been processed (as determined by the post-processor specified
    in its configuration)."""

    def __init__(self, config):
        self._logger = logging.getLogger("LQMT.Source.Directory")
        self.files_to_process = None

        if 'dirs' not in config:
            raise ConfigurationError("Missing required key: 'dirs' in section: 'Source.Directory'")
        self._dirs = config["dirs"]
        hasError = False
        for dirName in self._dirs:
            if not os.path.exists(dirName):
                self._logger.error('dir ({0}) is not a valid path'.format(dirName))
                hasError = True
        if hasError:
            raise ConfigurationError()

        if 'post_process' not in config:
            post_process = "move"
        else:
            post_process = config["post_process"]
            if post_process not in ["move", "delete", "track", "nothing"]:
                raise ConfigurationError(
                    "Invalid value for key: 'post_process' in section: 'Source.Directory': " + post_process)
        self._processedHandler = self._getProcessedHandler(post_process)

    def getFilesToProcess(self):
        self.files_to_process = FilesToProcess(self._dirs, self._processedHandler)
        return self.files_to_process

    def _getProcessedHandler(self, post_process):
        if post_process == "move":
            return processed.ProcessHandlerMove()
        elif post_process == "delete":
            return processed.ProcessHandlerDelete()
        elif post_process == "track":
            return processed.ProcessHandlerTrackFile()
        else:
            return processed.ProcessHandlerDoNothing()

    def processed(self, datafile):
        self._processedHandler.processed(datafile)

    def logStatistics(self, numAlerts):
        self._logger.info(
            "dirs: {0} NumDirs: {1} NumFiles: {2}".format(",".join(self._dirs), self.files_to_process._numDirs,
                                                          self.files_to_process._numFiles))


import logging
from lqmt.lqm.tool import ToolConfig
from lqmt.lqm.exceptions import ConfigurationError


class SysLogConfig(ToolConfig):
    """
       ToSyslog config variables:
        * = required
        *host - address of remote syslog address
        port - port that remote syslog server is listening on(defaults to 514)
        *protocol - protocol you want to communicate with(tcp/udp)
        *messageHead - Define what you want to have at the beginning of every message sent to syslog
        *messageFields - specify what fields you want extracted from the alerts and included in the messages
            note: Selected fields will be appended to end of message as so:
            ex: *messageHead* indicatorType="IPV4" indicator="192.168.1.1"
    """
    def __init__(self, configData, csvToolInfo, unhandledCSV):
        super().__init__(configData, csvToolInfo, unhandledCSV)
        self._logger = logging.getLogger("LQMT.SysLog.{0}".format(self.getName()))
        hasError = False
        self.port = 514  # default port for syslog is 514.

        if 'host' in configData:
            self.host = configData['host']
        else:
            self._logger.error("Host must be specified in the configuration")
            hasError = True

        if 'port' in configData:
            self.port = int(configData['port'])
        else:
            self._logger.error("Port not specified. Defaulting to port 514. ")

        if 'protocol' in configData:
            protocol = configData['protocol']
            if protocol == 'udp' or protocol == 'tcp':
                self.protocol = protocol
            else:
                self._logger.error("Invalid protocol: {0}".format(protocol))
                hasError = True
            self.protocol = protocol
        else:
            self._logger.error("Protocol must be specified in the configuration")
            hasError = True

        if 'messageHead' in configData:
            self.messageHead = configData['messageHead']
        else:
            self._logger.error("A message head must be specified in the configuration.")

        if 'messageFields' in configData:
            self.messageFields = configData['messageFields']
        else:
            self._logger.error("Message fields must be specified in the configuration.")

        if hasError:
            self.disable()
            raise ConfigurationError("Missing a required value in the user configuration for the to_syslog tool")



try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup
    from setuptools import find_packages

setup(
    name='lqmt',
    version='1.3.0',
    description='Flexible framework that allows automation to process cyber threat information and update endpoint defense tools',
    long_description='The Last Quarter Mile Toolset is flexible framework that allows automation to process cyber threat information (CTI) and update endpoint defense tools.',
    url='https://github.com/anl-cyberscience/LQMToolset/',
    author='The CFM Team',
    author_email='fedhelp@anl.gov',
    classifiers=[
        # See: https://pypi.python.org/pypi?%3Aaction=list_classifiers

        # How mature is this project? Common values are
        # Development Status :: 1 - Planning
        # Development Status :: 2 - Pre-Alpha
        # Development Status :: 3 - Alpha
        # Development Status :: 4 - Beta
        # Development Status :: 5 - Production/Stable
        # Development Status :: 6 - Mature
        # Development Status :: 7 - Inactive
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Information Technology',
        'Topic :: Security',

        # Pick your license as you wish (should match 'license' above)
        'License :: Other/Proprietary License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6'
    ],
    keywords='firewall endpoint last quarter mile toolset palo alto syslog checkpoint flextext',
    packages=find_packages(exclude=['contrib', 'doc', 'tests*']),
    install_requires=[
        'lxml',
        'netaddr',
        'pan-python',
        'toml',
        'FlexTransform',
        'requests'
    ],
    entry_points={
        'console_scripts': [
            'lqmt = lqmt.LQMT:main'
        ]
    },
    test_suite='nose.collector',
    tests_require=['nose']
)



from .whitelist import Whitelist

class HostWL(Whitelist):
    """Domain whitelist.
    This class stores its whitelisted hosts in the host table.  If a host is in the table it is whitelisted.
    """
    @staticmethod
    def getCreateTable():
        """Create the table to hold the hostnames"""
        return "create table host (host char(256))"

    @staticmethod
    def storeDB(conn,hosts):
        """Overwrite the hosts in the table with the specified hosts (string) in the database."""
        c=conn.cursor()
        vals=[]
        for host in hosts:
            vals.append( (host,) )
        c.execute("delete from host")
        c.executemany("insert into host(host) values (?)",iter(vals))
        c.close()

    def isWhitelisted(self,conn,indicatorType,indicator):
        """Return whether or not the indicator of type indicatorType is whitelisted by this whitelist.
        This checks the indicator against the hosts in the table and requires an exact match to be considered whitelisted."""
        c=conn.cursor()
        c.execute("select host from host where host=?",(indicator,))
        rec=c.fetchone()
        c.close()
        return rec != None

        


import ast
import logging
from lqmt.lqm.logging import LQMLogging
from .config import LQMToolConfig


# based on filename, place file either in the metafiles dict or the datafiles list
# NOTE: Assumes metafiles begin with . and datafiles do not.

class LQMToolController:
    def __init__(self, configfile):
        """
        Controller for LQMT.
        :param configfile: User configuration file
        """

        self._logger = logging.getLogger("LQMT.Controller")
        self._logger.info("Starting LQMTool")

        self._config = LQMToolConfig(configfile)
        self.toolChains = self._config.getToolChains()
        self.numAlerts = 0
        self.src = None

    def run(self):
        """
        Main function of the controller. Runs through the various methods used to gather alert files, parse them,
        and send the parsed alert data to the various tools.
        """

        alert_files = self._initialize()

        for data, unparsed_metadata in alert_files:
            metadata = self._parsemeta(unparsed_metadata)
            if metadata:
                self._parse(data, metadata)
        self._chainCleanup()

    def _parsemeta(self, metafile):
        """
        Parses metadata files
        :param metafile: Path to metadata file
        :return: Returns parsed metadata if the file path is valid. If not, then it returns None
        """

        try:
            f = open(metafile, 'r')
            meta = ast.literal_eval(f.read())
            f.close()
            return meta
        except Exception as inst:
            self._logger.error('An exception occurred while opening/parsing {0}:'.format(metafile))
            self._logger.error(str(inst))
            return None

    def _initialize(self):
        """
        Initializes LQMT. This includes initializing toolchains and their respective tools, and then getting all alert
        files specified from sources defined in the user configuration.
        :return: Returns filesToProcess object, which is defined in sourcedir.py. It's an object that has a custom
        __iter__ method that is used to traverse the top level alert directories provided by the user in user config
        """

        filesToProcess = None

        for chain in self.toolChains:
            if chain.isEnabled():
                chain.initialize()
            chain.updateEnabled()

        for src in self._config.getSources():
            self.src = src
            filesToProcess = src.getFilesToProcess()

        return filesToProcess

    def _parse(self, data, metadata):
        """
        Defines all parses needed based on the metadata given. Once defined, the parsers are used to parse alert data
        and pass the data to tools.
        :param data: Alert data
        :param metadata: Alert metadata
        """

        parser = self._config.getParser(metadata["PayloadFormat"])

        try:
            if parser is not None:
                # tell each chain there is a new file
                alerts = parser.parse(data, metadata)
                if alerts:
                    for chain in self.toolChains:
                        if chain.isEnabled():
                            chain.fileBegin()
                            self._process_alerts(alerts, chain, data, metadata)
                            self.src.processed(data)
                else:
                    self._logger.error("Processing error occurred. No processed alert data returned to LQMT.")
        except Exception as e:
            msg = "An exception occurred while processing file '{0}'".format(data)

            if not LQMLogging.isDebug():
                self._logger.error(msg)
            else:
                self._logger.exception(msg)
            self._logger.error(str(e))

    def _process_alerts(self, alerts, chain, datafile, metadata):
        """
        Takes alert data and passes it to active toolchains and their respective tools.
        :param alerts: Alert data
        :param chain: Toolchain
        """

        for alert in alerts:
            self.numAlerts += 1
            isWL = alert.isWhitelisted(self._config.getWhitelist())
            chain.process(alert, isWL, datafile, metadata)
            chain.fileDone()

    def _chainCleanup(self):
        """
        Cleans up tools that are done processing and logs statistics on amount of processed alerts.
        """

        for chain in self.toolChains:
            if chain.isEnabled():
                chain.commit()
                chain.cleanup()

        for src in self._config.getSources():
            src.logStatistics(self.numAlerts)